@@ -1,6 +1,6 @@
 package Net::Works::Address;
 {
-  $Net::Works::Address::VERSION = '0.01';
+  $Net::Works::Address::VERSION = '0.02';
 }
 BEGIN {
   $Net::Works::Address::AUTHORITY = 'cpan:DROLSKY';
@@ -10,9 +10,14 @@ use strict;
 use warnings;
 
 use Carp qw( confess );
-use NetAddr::IP::Util qw( inet_any2n );
-use NetAddr::IP;
+use Data::Validate::IP qw( is_ipv4 );
+use Math::BigInt try => 'GMP,Pari,FastCalc';
+use NetAddr::IP::Util qw( bin2bcd bcd2bin ipv6_n2x );
+use Net::Works::Types qw( IPInt PackedBinary Str );
 use Scalar::Util qw( blessed );
+use Socket qw( AF_INET AF_INET6 inet_pton inet_ntop );
+
+use integer;
 
 # Using this currently breaks overloading - see
 # https://rt.cpan.org/Ticket/Display.html?id=50938
@@ -20,69 +25,100 @@ use Scalar::Util qw( blessed );
 #use namespace::autoclean;
 
 use overload (
-    q{""} => 'as_string',
+    q{""} => '_overloaded_as_string',
     '<=>' => '_compare_overload',
 );
 
 use Moose;
 
-has _ip => (
-    is      => 'ro',
-    isa     => 'NetAddr::IP',
-    handles => {
-        version     => 'version',
-        mask_length => 'bits',
-    },
+with 'Net::Works::Role::IP';
+
+has _binary => (
+    is       => 'ro',
+    reader   => 'as_binary',
+    isa      => PackedBinary,
+    required => 1,
 );
 
-override BUILDARGS => sub {
-    my $class = shift;
+has _integer => (
+    is      => 'ro',
+    reader  => 'as_integer',
+    isa     => IPInt,
+    lazy    => 1,
+    builder => '_build_integer',
 
-    my $p = super();
+);
 
-    my $ip
-        = $p->{_ip} // ( $p->{version} && $p->{version} == 6 )
-        ? NetAddr::IP->new6( $p->{address} )
-        : NetAddr::IP->new( $p->{address} )
-        or die "Invalid address: $p->{address}";
-
-    return { _ip => $ip };
-};
+has _string => (
+    is      => 'ro',
+    reader  => 'as_string',
+    isa     => Str,
+    lazy    => 1,
+    builder => '_build_string',
+);
 
 sub new_from_string {
     my $class = shift;
     my %p     = @_;
 
-    return $class->new( address => delete $p{string}, %p );
+    my $str     = delete $p{string};
+    my $version = delete $p{version};
+
+    if ( is_ipv4($str) ) {
+        $version ||= 4;
+        $str = '::' . $str if $version == 6;
+    }
+    else {
+        $version ||= 6;
+    }
+
+    my $family = $version == 6 ? AF_INET6 : AF_INET;
+
+    return $class->new(
+        _binary => inet_pton( $family, $str ),
+        version   => $version,
+        %p,
+    );
 }
 
 sub new_from_integer {
     my $class = shift;
     my %p     = @_;
 
-    return $class->new( address => delete $p{integer}, %p );
+    my $int     = delete $p{integer};
+    my $version = delete $p{version};
+    $version ||= ref $int ? 6 : 4;
+
+    if ( $version == 4 && ref($int) ) {
+        $int = $int->numify;
+    }
+
+    my $packed = $version == 4 ? pack( N => $int ) : bcd2bin($int);
+
+    return $class->new(
+        _binary => $packed,
+        version => $version,
+        %p,
+    );
 }
 
-sub as_string {
+sub _build_string {
     my $self = shift;
 
-    return $self->version() == 6
-        ? lc $self->_ip()->short()
-        : $self->_ip()->addr();
+    return inet_ntop( $self->address_family(), $self->as_binary() );
 }
 
-sub as_integer {
-    my $self = shift;
-
-    return $self->version() == 4
-        ? ( scalar $self->_ip()->numeric() ) + 0
-        : scalar $self->_ip()->bigint();
+# overload passes extra arguments to this sub for some reason
+sub _overloaded_as_string {
+    return $_[0]->as_string();
 }
 
-sub as_binary {
+sub _build_integer {
     my $self = shift;
 
-    return inet_any2n( $self->as_string );
+    return $self->version == 4
+        ? unpack( N => $self->as_binary() )
+        : Math::BigInt->new( bin2bcd( $self->as_binary() ) );
 }
 
 sub as_ipv4_string {
@@ -110,20 +146,22 @@ sub as_bit_string {
         return sprintf( '%0128s', $bin );
     }
     else {
-        return sprintf( '%032b', $self->as_integer );
+        return sprintf( '%032b', $self->as_integer() );
     }
 }
 
+sub mask_length { $_[0]->bits() }
+sub bits { $_[0]->version() == 6 ? 128 : 32 }
+
 sub next_ip {
     my $self = shift;
 
-    my $bits = $self->mask_length();
     confess "$self is the last address in its range"
-        if $self->as_integer() == do { use bigint; ( 2**$bits - 1 ) };
+        if $self->as_integer() == $self->_max;
 
     return __PACKAGE__->new_from_integer(
         integer => $self->as_integer() + 1,
-        version => $self->version()
+        version => $self->version(),
     );
 }
 
@@ -170,7 +208,7 @@ Net::Works::Address - An object representing a single IP (4 or 6) address
 
 =head1 VERSION
 
-version 0.01
+version 0.02
 
 =head1 SYNOPSIS
 
@@ -233,8 +271,8 @@ you'll need to set the version explicitly to get an IPv6 address.
 
 =head2 $ip->as_string()
 
-Returns a string representation of the address like "1.2.3.4" or
-"ffff::a:1234".
+Returns a string representation of the address in the same format as
+inet_ntop, e.g., "1.2.3.4", "::1.2.3.4", or "ffff::a:1234".
 
 =head2 $ip->as_integer()
 
@@ -265,6 +303,11 @@ Returns a 4 or 6 to indicate whether this is an IPv4 or IPv6 address.
 Returns the mask length for the IP address, which is either 32 (IPv4) or 128
 (IPv6).
 
+=head2 $ip->bits()
+
+An alias for C<< $ip->mask_length() >>. This helps make addresses & network
+objects interchangeable in some cases.
+
 =head2 $ip->next_ip()
 
 Returns the numerically next IP, regardless of whether or not it's in the same
